/*
 * Sixth Data. Copyright ©2012-2018, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 */

package eu.svjatoslav.sixth.data.store.file;

import java.io.IOException;

public class MetaData {

    public static final int FILE_LOCATION_ENTRY_ALLOCATION_TABLE_START = 4 * 6;
    private static final int FORMAT_VERSION = 1;
    private static final int FILE_LOCATION_VERSION = 4 * 0;
    private static final int FILE_LOCATION_ENTRIES_TABLE_SIZE = 4 * 1;
    private static final int FILE_LOCATION_USED_ENTRIES_COUNT = 4 * 2;
    private static final int FILE_LOCATION_CURRENT_ENTRY = 4 * 3;
    private static final int FILE_LOCATION_CURRENT_LOCATION = 4 * 4; // LONG!!
    private final FileDataStore dataStore;
    private int entriesTableSize;
    private int currentEntry;
    private int usedEntriesCount;
    private long currentLocation;

    public MetaData(final FileDataStore dataStore) {
        this.dataStore = dataStore;
    }

    /**
     * @return address of the start of the allocated space
     */
    public synchronized long allocateStorageSpace(final int amountOfBytes) {
        final long locationStart = currentLocation;
        currentLocation += amountOfBytes;
        return locationStart;
    }

    public void decreaseUsedEntriesCount() {
        usedEntriesCount--;
    }

    public synchronized void ensureMinimumCurrentLocation(
            final long minimumLocation) {

        if (currentLocation < minimumLocation)
            currentLocation = minimumLocation;
    }

    public boolean entriesTableNeedsIncreasing() {
        final int reserveSize = (entriesTableSize / 10) + 1;

        return usedEntriesCount >= (entriesTableSize - reserveSize);

    }

    public long getEntriesStorageAreaStart() {
        return FILE_LOCATION_ENTRY_ALLOCATION_TABLE_START
                + (EntryRecord.ENTRY_RECORD_LENGTH * entriesTableSize);
    }

    public long getEntriesStorageAreaStart(
            final int hypotheticalEntriesTableSize) {
        return FILE_LOCATION_ENTRY_ALLOCATION_TABLE_START
                + (EntryRecord.ENTRY_RECORD_LENGTH * hypotheticalEntriesTableSize);
    }

    /**
     * @return amount of entries in the entries table
     */
    public int getEntriesTableSize() {
        return entriesTableSize;
    }

    public void setEntriesTableSize(final int newSize) {
        entriesTableSize = newSize;
    }

    public int getNewEntryId() {
        currentEntry++;

        if (currentEntry >= getEntriesTableSize())
            currentEntry = 0;

        return currentEntry;
    }

    public void increaseUsedEntriesCount() {
        usedEntriesCount++;
    }

    public void initializeNewFile() throws IOException {
        entriesTableSize = 16;
        usedEntriesCount = 0;
        currentEntry = entriesTableSize - 1;
        currentLocation = getEntriesStorageAreaStart();

        writeFileHeader();
    }

    public void readFileHeader() throws IOException {

        final int fileVersion = dataStore.readInt(FILE_LOCATION_VERSION);
        if (fileVersion != FORMAT_VERSION)
            throw new RuntimeException("File version is " + fileVersion
                    + " but version " + FORMAT_VERSION + " is required.");

        entriesTableSize = dataStore.readInt(FILE_LOCATION_ENTRIES_TABLE_SIZE);

        usedEntriesCount = dataStore.readInt(FILE_LOCATION_USED_ENTRIES_COUNT);

        currentEntry = dataStore.readInt(FILE_LOCATION_CURRENT_ENTRY);

        currentLocation = dataStore.readLong(FILE_LOCATION_CURRENT_LOCATION);
    }

    public void setCurrentLocation(final long currentLocation) {
        this.currentLocation = currentLocation;
    }

    public void writeFileHeader() throws IOException {
        dataStore.writeInt(FILE_LOCATION_VERSION, FORMAT_VERSION);
        dataStore.writeInt(FILE_LOCATION_ENTRIES_TABLE_SIZE,
                getEntriesTableSize());
        dataStore.writeInt(FILE_LOCATION_USED_ENTRIES_COUNT, usedEntriesCount);
        dataStore.writeInt(FILE_LOCATION_CURRENT_ENTRY, currentEntry);
        dataStore.writeLong(FILE_LOCATION_CURRENT_LOCATION, currentLocation);
    }

}
